import copy
import datetime
import itertools
import os
import re
import sys
import time

import dateutil
tzinfos = {"GMT": dateutil.tz.gettz("GMT")}
import numpy as np
import pandas as pd
import json

from crossnetwork_processing_functions import get_data_files, create_GHOST_stage_logfile
from GHOST_standards import get_standard_data, get_standard_metadata, country_ISO3166_to_name, standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code
from network_file_standardise import NetworkFileStandardise
from network_parsers import ParseUSNADPFile
from standardisation_mapping_functions import map_network_classifications_to_standard_classifications, map_network_measurement_information_to_standard_measurement_information, network_data_flag_to_standard_data_flag_mapping
from time_functions import round_to_minute, datetime_to_fractional_minutes_from_reference, get_datafile_download_timestamp
import unit_converter


class StandardiseNetworkData:

    '''Convert 1 provided data file to GHOST standard numpy station specific files'''

    def __init__(self, null, run_details, parameter_details, network_details):
        
        #add passed variables to self
        self.run_details = run_details
        self.parameter_details = parameter_details
        self.network_details = network_details

    def do_work(self):
        '''Function which does executes the functions of the class in a pipeline'''

        #get data file to parse
        data_file = self.gather_data_files()

        #parse station reference data files
        self.parse_file(data_file)

        #write a valid process log file
        self.run_details['log_file_str'] += str((time.time() - self.run_details['stage_process_start_time'])/60.)
        create_GHOST_stage_logfile(0,self.run_details)


    def gather_data_files(self):

        '''Function to gather data files to parse per station'''

        #get data files specific to station
        data_file = get_data_files(self.run_details, self.parameter_details)[0]   

        #load external metadata into memory 
        with open(self.run_details['data_root']+'US_NADP_MDN/metadata/{}/processed/US_NADP_MDN_META.json'.format(self.run_details['ghost_version']), 'r', encoding='utf-8') as f:
            self.external_metadata = json.loads(f.read())
        #self.external_metadata = np.load(self.run_details['data_root']+'US_NADP_MDN/metadata/processed/US_NADP_MDN_META.json', allow_pickle=True)

        #get standard measured unit quantity
        conv_obj = unit_converter.convert_units(self.parameter_details['standard_units'],self.parameter_details['standard_units'],1)
        self.measured_unit_quantity = conv_obj.output_represented_quantity

        return data_file

    def parse_file(self, data_file):

        '''Function to parse data file to GHOST standard numpy station specific file'''

        self.run_details['log_file_str'] += 'PROCESSING FILE: {}\n'.format(data_file) 

        #create object for parsing US NADP MDN annual file, and read file
        parsing_obj = ParseUSNADPFile(data_file)
        parsing_obj.read_data()

        #get unique station references
        unique_station_references = np.unique(parsing_obj.parsed_data['siteID']) 

        #iterate through unique station references
        for unique_station_reference in unique_station_references:

            if unique_station_reference in list(self.external_metadata.keys()): # filter stations that appear in metadata

                self.run_details['log_file_str'] += '{} -- PARSING RAW DATA\n'.format(unique_station_reference)

                #create site instances of metadata/data dictionaries
                site_data =     get_standard_data(self.parameter_details)
                site_metadata = get_standard_metadata(self.parameter_details)

                #isolate variable data specific for station 
                valid_inds = parsing_obj.parsed_data['siteID'] == unique_station_reference
                start_dt = parsing_obj.parsed_data['dateOn'][valid_inds]
                end_dt = parsing_obj.parsed_data['dateOff'][valid_inds]
                measurements = parsing_obj.parsed_data[self.run_details['parameter']][valid_inds]
                qr = parsing_obj.parsed_data['qr'][valid_inds]
                notes = parsing_obj.parsed_data['notes'][valid_inds]

                #------------------------------------------------------------------------#
                #  PARSE DATA per station
                #------------------------------------------------------------------------#
                
                valid_start_dt = start_dt
                valid_end_dt = end_dt
                valid_measurements = measurements
                valid_qr = qr
                valid_notes = notes

                #if have no valid measurements saved after all that continue to next station
                if len(valid_measurements) == 0:
                    self.run_details['log_file_str'] += '{} -- NO VALID MEASUREMENTS\n'.format(unique_station_reference)
                    continue
                
                #------------------------------------------------------------------------#
                #PARSE METADATA
                #------------------------------------------------------------------------#

                #GHOST version
                site_metadata['GHOST_version']['value'].append(self.run_details['ghost_version'])

                #initialise string which will store all GHOST process warnings as empty            
                site_metadata['process_warnings']['value'].append('')

                #write known metadata  
                site_metadata['network']['value'].append(self.run_details['network'])
                site_metadata['data_level']['value'].append('none')  
                site_metadata['contact_name']['value'].append('David Gay')
                site_metadata['contact_institution']['value'].append('National Atmospheric Deposition Program')      
                site_metadata['contact_email_address']['value'].append('David.Gay@slh.wisc.edu')    
                site_metadata['principal_investigator_name']['value'].append('David Gay')
                site_metadata['principal_investigator_institution']['value'].append('National Atmospheric Deposition Program')      
                site_metadata['principal_investigator_email_address']['value'].append('David.Gay@slh.wisc.edu')  
                site_metadata['data_licence']['value'].append('Public domain: https://edg.epa.gov/EPA_Data_License.html')   
                #assume datum information
                site_metadata['ellipsoid']['value'].append('WGS 84')
                site_metadata['horizontal_datum']['value'].append('World Geodetic System 1984')
                site_metadata['vertical_datum']['value'].append('tidal - mean sea level')
                site_metadata['projection']['value'].append('geographic')   
                #sampling height is known to be 2 metres
                site_metadata['sampling_height']['value'].append('1.0')

                #add reported measurement unit
                #convert raw reported units given to standard name
                if self.run_details['parameter'] == 'PREC_ACCUM':
                    reported_units = 'mm'
                else:
                    reported_units = 'ng/l' # concentration in precip

                conv_obj = unit_converter.convert_units(reported_units, reported_units, 1, measured_species = self.parameter_details['chemical_formula'])
                site_metadata['measuring_instrument_reported_units']['value'].append(conv_obj.output_standard_units)

                # add external metadata
                cut_external_metadata = self.external_metadata[unique_station_reference]
                site_metadata['longitude']['value'].append(cut_external_metadata['longitude']['values'][-1])
                site_metadata['latitude']['value'].append(cut_external_metadata['latitude']['values'][-1])
                site_metadata['altitude']['value'].append(cut_external_metadata['elevation']['values'][-1])
                site_metadata['station_name']['value'].append(cut_external_metadata['siteName']['values'][-1])
            
                #parse station classifications (from external metadata)
                native_classifications = {'area_classification':cut_external_metadata['siteClass']['values'][-1]}
                site_metadata = map_network_classifications_to_standard_classifications(native_classifications, site_metadata, self.run_details)
            

                #process measurement process information
                site_metadata, measurement_qa_flags, standard_method_abbreviation = map_network_measurement_information_to_standard_measurement_information('', site_metadata, self.run_details, self.parameter_details)
                #set measurement qa codes which are True as joined metadata str
                site_metadata['measurement_qa_flags'] = {'value':[], 'data_type':object, 'string_format':'lower_short'}
                measurement_qa_flag_codes = []
                for standard_flag_name in measurement_qa_flags.keys():
                    if measurement_qa_flags[standard_flag_name]:
                        measurement_qa_flag_codes.append(str(standard_QA_name_to_QA_code[standard_flag_name]))
                site_metadata['measurement_qa_flags']['value'].append(','.join(measurement_qa_flag_codes))

                #iterate through metadata variables, if any columns have not been appended to add a NaN
                for column_header in list(site_metadata.keys()):
                    n_column = len(site_metadata[column_header]['value'])
                    #if column is empty, then append NaN
                    if n_column < 1:
                        site_metadata[column_header]['value'].append(np.NaN)
        
                #set station_reference (provided network ID + standard measurement method abbreviation)
                site_metadata['station_reference']['value'][-1] = '{}_{}'.format(unique_station_reference,standard_method_abbreviation) 

                #get data file download timestamp
                site_metadata['data_download_stamp']['value'][-1] = get_datafile_download_timestamp(data_file)

                #add data revision timestamp, set as 0001-01-01 00:00 for all files
                data_revision_timestamp = datetime_to_fractional_minutes_from_reference(datetime.datetime(1,1,1,0,0))
                site_metadata['data_revision_stamp']['value'][-1] = data_revision_timestamp

                #------------------------------------------------------------------------#
                #PARSE TIME
                #------------------------------------------------------------------------#
            
                #parse start, end and modification datetime for each reported timepoint 
                #first create a unified date/time string (yyyy-mm-dd hh:mm:ss Z, of the measurement --> z is always GMT)
                #secondly parse into datetime (taking into account timezone information). then convert to UTC datetime, then round to nearest minute    
                start_dt = [round_to_minute(dateutil.parser.parse('{} GMT'.format(date), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in valid_start_dt] 
                end_dt = [round_to_minute(dateutil.parser.parse('{} GMT'.format(date), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in valid_end_dt] 
            
                #convert each start measurement datetime to minutes from reference time
                frac_time = [datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]

                #get timedeltas between start and end of measurement window
                temporal_resolutions = []
                for start, end in zip(start_dt,end_dt):
                    dt_window = end - start
                    #convert hours to minutes
                    dt_window_minutes = int(dt_window.total_seconds()/60.)
                    temporal_resolutions.append(dt_window_minutes) 

                #save meta_update_stamp, set as first sorted time processed (as minutes relative to reference time)
                site_metadata['meta_update_stamp']['value'][-1] = sorted(frac_time)[0]

                #------------------------------------------------------------------------#
                #PARSE QA/FLAGS
                #------------------------------------------------------------------------#

                flags = []
                qa = []

                #get dictionary necessary for mapping of US NADP MDN file data flags to standard data flag codes
                data_flag_mapping_dictionary = network_data_flag_to_standard_data_flag_mapping['US_NADP_MDN']                

                #iterate through network qr flags and note flags
                for qr_flag, note_flags in zip(valid_qr, valid_notes):

                    #lists to append qa/data flags per measurement
                    flags_per_measurement = []
                    qa_per_measurement = []
                    #set GHOST decreed validity/network decreed validity variables initially as 'V'
                    ghost_decreed_validity = 'V'
                    network_decreed_validity = 'V'

                    #split apart note flags
                    note_flags = list(note_flags) 

                    #do you have notes? if not use qr flag for mapping
                    if len(note_flags) > 0:

                        #iterate through all notes:
                        for note_flag in note_flags:

                            # map note flag and append
                            standard_data_flag_dict = data_flag_mapping_dictionary[note_flag]
                            #get mapped standard data flags
                            standard_data_flag_names = standard_data_flag_dict['standard_data_flag_name']
                            #iterate through standard data flags 
                            for standard_data_flag_ii, standard_data_flag_name in enumerate(standard_data_flag_names):
                                #check if standard data flag is empty (i.e. unmapped as non-useful)
                                if standard_data_flag_name != '': 
                                    #get corresponding data flag code for standard data flag
                                    flag_code = standard_data_flag_name_to_data_flag_code[standard_data_flag_name]
                                    #append standard data flag (if not already appended)
                                    if flag_code not in flags_per_measurement:
                                        flags_per_measurement.append(flag_code)
                                    #determine if GHOST/network decreed valid
                                    if standard_data_flag_dict['GHOST_decreed_validity'][standard_data_flag_ii] == 'I':
                                        ghost_decreed_validity = 'I'
                                    if standard_data_flag_dict['network_decreed_validity'][standard_data_flag_ii] == 'I':
                                        network_decreed_validity = 'I'
                
                    #use qr flag for mapping
                    else:                    
                        # map qr flag and append
                        standard_data_flag_dict = data_flag_mapping_dictionary[qr_flag]
                        #get mapped standard data flags
                        standard_data_flag_names = standard_data_flag_dict['standard_data_flag_name']
                        #iterate through standard data flags 
                        for standard_data_flag_ii, standard_data_flag_name in enumerate(standard_data_flag_names):
                            #check if standard data flag is empty (i.e. unmapped as non-useful)
                            if standard_data_flag_name != '': 
                                #get corresponding data flag code for standard data flag
                                flag_code = standard_data_flag_name_to_data_flag_code[standard_data_flag_name]
                                #append standard data flag (if not already appended)
                                if flag_code not in flags_per_measurement:
                                    flags_per_measurement.append(flag_code)
                                #determine if GHOST/network decreed valid
                                if standard_data_flag_dict['GHOST_decreed_validity'][standard_data_flag_ii] == 'I':
                                    ghost_decreed_validity = 'I'
                                if standard_data_flag_dict['network_decreed_validity'][standard_data_flag_ii] == 'I':
                                    network_decreed_validity = 'I'

        
                    #if ghost decreed validity, or network decreed validity is Invalid, then set QA flags
                    if ghost_decreed_validity == 'I':
                        qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
                    if network_decreed_validity == 'I':
                        qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
            
                    #append qa_per_measurement, flags_per_measurement to lists to write out
                    flags.append(flags_per_measurement)
                    qa.append(qa_per_measurement)

                #------------------------------------------------------------------------#
                #PARSE MEASUREMENTS
                #------------------------------------------------------------------------#

                #measurement values need converting (i.e. mass density to mixing ratio)

                #measurement temperature and pressure values are never provided and are seemingly not measured as the measurements are passive.
                #however measurements are provided in mass density and need converting to mixing ratio, so do this using the US EPA standard pressure and temperatures
                #these are defined as for 298.15 K temperature, and 1013.25 hPa for pressure here for all compounds, except PM2.5 and lead: https://www.govinfo.gov/content/pkg/CFR-2011-title40-vol2/pdf/CFR-2011-title40-vol2-chapI.pdf

                measurements = []

                site_metadata['network_provided_volume_standard_temperature']['value'][-1] = np.NaN
                site_metadata['network_provided_volume_standard_pressure']['value'][-1] = np.NaN
                
                #get standard reported units/quantity
                standard_reported_units = conv_obj.output_standard_units
                measuring_instrument_reported_unit_quantity = conv_obj.output_represented_quantity  

                #set flags for writing QA flags associated with assumptions in unit conversion
                unit_conversion_standard_network = False
                unit_conversion_educated_guess = False

                #check if reported standard units, match required standard units
                if standard_reported_units != site_data[self.parameter_details['bsc_parameter_name']]['units']:

                    #only do simple unit conversion, i.e. pptv --> ppbv, ng/m3 --> ug/m3
                    conv_obj = unit_converter.convert_units(standard_reported_units,site_data[self.parameter_details['bsc_parameter_name']]['units'],1.0, measured_species = self.parameter_details['chemical_formula'])

                    #process through each measurement and convert value using derived conversion factor
                    for n, meas in enumerate(valid_measurements):
                        #only do conversion if measurement value is above 0 and not infinite
                        try:
                            float_meas = float(meas)
                        except:
                            float_meas = np.NaN
                        
                        if (float_meas > 0.0) & (np.isfinite(float_meas)) & (standard_data_flag_name_to_data_flag_code['Missing Data'] not in flags[n]):
                            measurements.append(float_meas*conv_obj.conversion_factor)
                        else:
                            measurements.append(float_meas)

                #otherwise do nothing to measured values
                else:
                    for n, meas in enumerate(valid_measurements):
                        try:
                            float_meas = float(meas)
                        except:
                            float_meas = np.NaN
                        measurements.append(float_meas)

                #keep only measurement windows that are >= 0 minute in temporal resolution, as well as all other associated variables across other fields
                valid_temporal_resolution_inds = [ii for ii, res in enumerate(temporal_resolutions) if res >= 0]
                if len(valid_temporal_resolution_inds) != len(frac_time):
                    frac_time = [frac_time[x] for x in valid_temporal_resolution_inds]
                    temporal_resolutions = [temporal_resolutions[x] for x in valid_temporal_resolution_inds]
                    flags = [flags[x] for x in valid_temporal_resolution_inds]
                    qa = [qa[x] for x in valid_temporal_resolution_inds]
                    measurements = [measurements[x] for x in valid_temporal_resolution_inds]

                #----------------------------------------------------------------------------#
                #WRITE OUT DATA
                #----------------------------------------------------------------------------# 
                site_data['time']['value'] = frac_time
                site_data[self.parameter_details['bsc_parameter_name']]['value'] = measurements
                site_data['temporal_resolution']['value'] = temporal_resolutions   
                site_data['flag']['value'] = flags
                site_data['qa']['value'] = qa
                #iterate through data variables, if any columns have not been appended to, add a list of NaNs (equal to length of time array)
                for column_header in list(site_data.keys()):
                    n_column = len(site_data[column_header]['value'])
                    #if column is empty, then append NaN
                    if n_column == 0:
                        if site_data[column_header]['data_type'] != np.uint8:
                            site_data[column_header]['value'] = [np.NaN]*len(site_data['time']['value'])
                        else:
                            site_data[column_header]['value'] = [255]*len(site_data['time']['value'])
                #------------------------------------------------------

                #standardise metadata/data formats and write out station data to numpy file
                write_obj = NetworkFileStandardise(site_metadata, site_data, self.run_details, self.parameter_details)
                if write_obj.written == False:
                    self.run_details['log_file_str'] += 'DATA ALL MISSING FOR STATION: {}. SKIPPING.\n'.format(unique_station_reference)